// Copyright (c) 2014, The Shelf Bind project authors.
// Please see the AUTHORS file for details.
// All rights reserved. Use of this source code is governed by
// a BSD 2-Clause License that can be found in the LICENSE file.

library shelf_bind.bindings;

import 'package:shelf/shelf.dart';
import 'package:shelf_path/shelf_path.dart';
import 'dart:mirrors';
import 'preconditions.dart';
import 'util.dart';
import 'annotations.dart';
import 'converters.dart';
import 'dart:async';
import 'dart:convert';
import 'package:quiver/collection.dart';
import 'package:path/path.dart' as p;
import 'dart:io';

final _url = p.url;

HandlerBinding createHandlerBinding(Function handler,
    List<HandlerParameterBinding> providedBindings,
    ResponseBinding responseBinding) {
  return _inferHandlerBindings(handler, providedBindings,
      responseBinding);
}


HandlerParameterBinding createRequestParameterBinding(
    Type type, Symbol constructor, Map<String, Symbol> bindings,
    BindMode bindMode) {
  return new TypePathParameterBinding(
      new TypeBinding(type, constructorName: constructor,
          bindMode: bindMode, memberBindings: bindings));
}

HandlerParameterBinding createJsonBodyBinding(Type type) {
  return new TypeBodyParameterBinding(new TypeBinding(type));
}


class BindMode {
  final String literal;

  const BindMode._internal(this.literal);

  static const BindMode CONSTRUCTOR = const BindMode._internal("CONSTRUCTOR");
  static const BindMode PROPERTY = const BindMode._internal("PROPERTY");
  static const BindMode CONSTRUCTOR_THEN_PROPERTY =
      const BindMode._internal("CONSTRUCTOR_THEN_PROPERTY");
}

class HandlerBinding {
  final List<HandlerParameterBinding> parameterBindings;
  final Function _boundHandler;
  final ResponseBinding responseBinding;

  HandlerBinding(this.parameterBindings, Function handler,
      [ ResponseBinding responseBinding ])
      : this._boundHandler = _wrapHandler(handler),
        this.responseBinding = responseBinding != null ?
            responseBinding : new JsonResponseBinding.ok() {
    ensure(parameterBindings, isNotNull);
    ensure(_boundHandler, isNotNull);
    parameterBindings.forEach((pb) => pb._validate());
  }


  Handler get handler => _handleRequest;

  dynamic _handleRequest(Request request) {
    final parameterFutureValues =
        parameterBindings.map((pb) => syncFuture(() =>
            pb.createInstance(request)));

    return Future.wait(parameterFutureValues)
      .then((parameterValues) {
        return _boundHandler(parameterValues, request, responseBinding);
      });

  }


}

/**
 * A binding from instances of [Request] to a single parameter of a handler
 * function.
 *
 * Different sorts of bindings are supported such as:
 * * the [request] object itself (see [RequestObjectParameterBinding])
 * * path parameters to simple type (see [SimplePathParameterBinding])
 * * path parameters to class types (see [TypePathParameterBinding])
 * * request body in JSON format to class types (see [TypeBodyParameterBinding])
 */
abstract class HandlerParameterBinding {
  dynamic createInstance(Request request);
  void _validate();
}

abstract class _TypeParameterBinding extends HandlerParameterBinding {
  final TypeBinding typeBinding;

  _TypeParameterBinding(this.typeBinding);

  @override
  void _validate()  {
    ensure(typeBinding, isNotNull);
    typeBinding._validateBindings();
  }

}

class RequestObjectParameterBinding extends HandlerParameterBinding {

  @override
  void _validate() {}

  @override
  createInstance(Request request) => request;
}

class SimplePathParameterBinding extends HandlerParameterBinding {
  final String pathParameterName;
  final Type simpleType;

  SimplePathParameterBinding(this.pathParameterName,
      [this.simpleType = String]);

  @override
  dynamic createInstance(Request request) {
    // TODO: simple type conversion
    final value = getPathParameter(request, pathParameterName);
    return value == null || simpleType == String ?
        value : convertTo(simpleType, value);
  }

  @override
  void _validate() {
    ensure(pathParameterName, isNotNull);
  }
}

class TypePathParameterBinding extends _TypeParameterBinding {
  TypePathParameterBinding(TypeBinding typeBinding) : super(typeBinding);


  @override
  dynamic createInstance(Request request) {
    final instanceMirror = _createInstance(request, typeBinding);
    return instanceMirror.reflectee;
  }

}

class TypeBodyParameterBinding extends _TypeParameterBinding {
  TypeBodyParameterBinding(TypeBinding typeBinding) : super(typeBinding);

  @override
  dynamic createInstance(Request request) {
    final jsonFuture = _extractJsonBody(request);
    return jsonFuture.then((json) {
      final classMirror = reflectClass(typeBinding.type);
      final instanceMirror =
          classMirror.newInstance(#fromJson, [json]);

      return instanceMirror.reflectee;
    });
  }

  Future<Map> _extractJsonBody(Request request) =>
      request.readAsString().then((s) => JSON.decode(s));

}


class TypeBinding {
  final Type type;
  final Symbol constructorName;
  final Map<Symbol, Symbol> memberBindings;
  final BindMode bindMode;

  TypeBinding._internal(this.type, this.constructorName,
      this.bindMode, this.memberBindings) {
//    _validateBindings(); TODO: don't like turning this off
  }

  TypeBinding(Type type, {Symbol constructorName: #build,
      BindMode bindMode: BindMode.CONSTRUCTOR_THEN_PROPERTY,
      Map<String, Symbol> memberBindings: const {} })
      : this._internal(type, constructorName, bindMode,
          _convertBindings(memberBindings));

  static Map<Symbol, Symbol> _convertBindings(
      Map<String, Symbol> memberBindings) {

    if (memberBindings == null) {
      return const {};
    }

    final convertedBindings = <Symbol, Symbol>{};
    memberBindings.forEach((n, s) {
      convertedBindings[new Symbol(n)] = s;
    });
    return convertedBindings;
  }

  _validateBindings() {
    ensure(type, isNotNull);
    ensure(constructorName, isNotNull);
    ensure(memberBindings, isNotNull);
    ensure(memberBindings, isNot(isEmpty));
    ensure(bindMode, isNotNull);
  }
}

abstract class ResponseBinding {
  const ResponseBinding();

  createResponse(Request request, InstanceMirror responseMirror);
  void _validate();
}

abstract class BaseResponseBinding extends ResponseBinding {
  final int successStatus;
  final bool setLocation;

  const BaseResponseBinding(this.successStatus, {this.setLocation: false});
  const BaseResponseBinding.ok() : this(HttpStatus.OK);
  const BaseResponseBinding.created() : this(HttpStatus.CREATED,
      setLocation: true);

  createResponse(Request request, InstanceMirror responseMirror) {
    final r = responseMirror.reflectee;
    if (r == null) {
      return new Response.internalServerError(body:
        'handler method returned null');
    }

    if (r is Response) {
      return r;
    }

    return doCreateResponse(request, responseMirror);
  }

  Response doCreateResponse(Request request, InstanceMirror responseMirror) {
    final body = createBody(request, responseMirror);
    final statusCode = createStatusCode(request, responseMirror, body);
    final headers = createHeaders(request, responseMirror, body);
    return new Response(statusCode, body: body, headers: headers);
  }

  createBody(Request request, InstanceMirror responseMirror);

  int createStatusCode(Request request, InstanceMirror responseMirror,
                       dynamic body) => successStatus;

  Map<String, String> createHeaders(Request request,
      InstanceMirror responseMirror, dynamic body) {
    final headers = {};
    if (setLocation) {
      final location = _getLocation(request, responseMirror);
      headers['Location'] = location;
    }
    addHeaders(headers, request, responseMirror, body);
    return headers;
  }

  String _getLocation(Request request, InstanceMirror responseMirror) {
    if (request.method != 'POST') {
      return request.requestedUri.toString();
    }

    final String id = getId(request, responseMirror);
    final requestPath = _url.fromUri(request.requestedUri);
    return _url.join(requestPath, id);
  }

  String getId(Request request, InstanceMirror responseMirror) {
    // TODO: support annotation like @CreatedId(#identifier)
    final idSymbol = #id;

    final id = responseMirror.getField(idSymbol);
    return id.reflectee;
  }

  void addHeaders(Map<String, String> headers,
                  Request request, InstanceMirror responseMirror,
                  dynamic body) {}

  @override
  void _validate() {
    ensure(successStatus, isNotNull);
  }
}

class JsonResponseBinding extends BaseResponseBinding {

  const JsonResponseBinding(int successStatus, {bool setLocation: false})
      : super(successStatus, setLocation: setLocation);

  const JsonResponseBinding.ok() : super.ok();

  const JsonResponseBinding.created() : super.created();

  @override
  createBody(Request request, InstanceMirror responseMirror) =>
      JSON.encode(responseMirror.reflectee);

  @override
  void addHeaders(Map<String, String> headers,
                  Request request, InstanceMirror responseMirror,
                  dynamic body) {
      headers['Content-Type'] = 'application/json';
  }

}



typedef _BoundHandler(List<dynamic> positionalArguments, Request request,
    ResponseBinding responseBinding);

HandlerBinding _inferHandlerBindings(Function handler,
    List<HandlerParameterBinding> providedParameterBindings,
    ResponseBinding providedResponseBinding) {
  final ClosureMirror functionMirror = reflect(handler);
  final inferredParameterBindings =
      _inferHandlerParameterBindings(functionMirror, providedParameterBindings);

  final inferredResponseBinding = _inferResponseBinding(
      functionMirror, providedResponseBinding);
  return new HandlerBinding(inferredParameterBindings, handler,
      inferredResponseBinding);
}

ResponseBinding _inferResponseBinding(ClosureMirror functionMirror,
                                      ResponseBinding providedResponseBinding) {
  if (providedResponseBinding != null) {
    return providedResponseBinding; // TODO: should probably merge
  }

  ResponseHeaders responseHeaders =
      _getAnnotation(functionMirror.function, ResponseHeaders);

  return responseHeaders != null ?
      new JsonResponseBinding(responseHeaders.successStatus,
          setLocation: responseHeaders.setLocation) :
            new JsonResponseBinding.ok();
}

List<HandlerParameterBinding> _inferHandlerParameterBindings(
    ClosureMirror functionMirror,
    List<HandlerParameterBinding> providedBindings) {
  final params = functionMirror.function.parameters;
  final inferredBindings = <HandlerParameterBinding>[];

  for (int i = 0; i < params.length; i++) {
    final param = params[i];
    final providedBinding = i > providedBindings.length - 1 ?
        null : providedBindings.elementAt(i);
    final parameterBinding = _inferHandlerParameterBinding(providedBinding,
        param);
    if (parameterBinding != null) {
      inferredBindings.add(parameterBinding);
    }
  }
  return inferredBindings;
}

HandlerParameterBinding _inferHandlerParameterBinding(
    HandlerParameterBinding providedBinding, ParameterMirror param) {
  if (param.type.reflectedType == Request) {
    return new RequestObjectParameterBinding();
  }
  if (_isSimpleType(param.type)) {
    return _inferSimplePathParameterBinding(providedBinding, param);
  }
  else {
    return _inferTypeParameterBinding(providedBinding, param);
  }
}

HandlerParameterBinding _inferSimplePathParameterBinding(
    HandlerParameterBinding providedBinding, ParameterMirror param) {

  if (providedBinding != null && providedBinding is _TypeParameterBinding) {
    throw new ArgumentError('type bindings not supported for dart core classes');
  }

  final functionParameterName = MirrorSystem.getName(param.simpleName);
  final providedName = providedBinding == null ? null :
    (providedBinding as SimplePathParameterBinding).pathParameterName;

  final pathParameterName = providedName != null ?
      providedName : functionParameterName;

  return new SimplePathParameterBinding(pathParameterName, param.type.reflectedType);

}

HandlerParameterBinding _inferTypeParameterBinding(
    HandlerParameterBinding providedBinding, ParameterMirror param) {

  if (providedBinding != null && providedBinding is SimplePathParameterBinding) {
    throw new ArgumentError(
        'simple bindings not supported for non dart core classes');
  }

  providedBinding = _inferTypeParameterBindingFromAnnotation(providedBinding, param);

  final providedTypeBinding = providedBinding == null ? new TypeBinding(null)
      : (providedBinding as _TypeParameterBinding).typeBinding;

  final inferredTypeBinding = _inferTypeBinding(providedTypeBinding, param);

  final inferredParameterBinding =
      (providedBinding == null || providedBinding is TypePathParameterBinding)
      ? new TypePathParameterBinding(inferredTypeBinding)
      : new TypeBodyParameterBinding(inferredTypeBinding);

  inferredParameterBinding._validate();
  return inferredParameterBinding;
}


HandlerParameterBinding _inferTypeParameterBindingFromAnnotation(
    HandlerParameterBinding providedBinding, ParameterMirror param) {

  final RequestBody requestBodyAnnotation = _getAnnotation(param, RequestBody);

  if (requestBodyAnnotation == null) {
    return providedBinding;
  }

  if (requestBodyAnnotation.format != BodyFormat.JSON) {
    throw new ArgumentError("Only Json bodies currently supported");
  }

  final TypeBinding typeBinding = new TypeBinding(requestBodyAnnotation.type);
  return new TypeBodyParameterBinding(typeBinding);


}

List<TypeBinding> _inferMissingBindings(Function handler,
    List<TypeBinding> typeBindings) {

  final ClosureMirror functionMirror = reflect(handler);
  final params = functionMirror.function.parameters;
  final inferredBindings = <TypeBinding>[];

  for(int i = 0; i < params.length; i++) {
    final param = params[i];
    final providedBinding = i > typeBindings.length - 1 ?
        new TypeBinding(null) : typeBindings.elementAt(i);
    final typeBinding = _inferTypeBinding(providedBinding, param);
    if (typeBinding != null) {
      inferredBindings.add(typeBinding);
    }
  }

  return inferredBindings;
}

// TODO: Not a good way to determine this
bool _isSimpleType(TypeMirror tm) =>
    (tm.isTopLevel && tm.owner.simpleName == #dart.core);


TypeBinding _inferTypeBinding(TypeBinding providedBinding, ParameterMirror pm) {
  final parameterType = pm.type;

  final constructorName = providedBinding.constructorName;
  final bindMode = providedBinding.bindMode;

  final typeMirror = _inferType(providedBinding.type, parameterType);
  if (!(typeMirror is ClassMirror)) {
    return null;
  }

  final cm = typeMirror;

  if (cm.isTopLevel && cm.owner.simpleName == #dart.core) {
    return null;
  }

  if (cm.reflectedType == Request) {
    return null;
  }

  final type = typeMirror.reflectedType;

  final bindModeAndBindings = _inferMemberBindings(
      providedBinding.memberBindings, typeMirror,
      bindMode, constructorName);

  return new TypeBinding._internal(type, constructorName,
      bindModeAndBindings.bindMode,
      bindModeAndBindings.memberBindings);
}

TypeMirror _inferType(Type providedType, TypeMirror parameterType) {
  return providedType != null ? reflectType(providedType) : parameterType;
}

class _ModeBindingPair {
  final BindMode bindMode;
  final Map<Symbol, Symbol> memberBindings;

  _ModeBindingPair(this.bindMode, this.memberBindings);
}

_ModeBindingPair _inferMemberBindings(
    Map<Symbol, Symbol> providedBindings, ClassMirror parameterType,
    BindMode bindMode, Symbol constructorName) {
  switch (bindMode) {
    case BindMode.CONSTRUCTOR_THEN_PROPERTY:
      return _inferBindingsViaConstructorThenProperty(providedBindings,
          parameterType, constructorName);
    case BindMode.CONSTRUCTOR:
      return new _ModeBindingPair(BindMode.CONSTRUCTOR,
          _inferMemberBindingsViaConstructor(providedBindings,
          parameterType, constructorName));
    case BindMode.PROPERTY:
      return new _ModeBindingPair(BindMode.PROPERTY,
          _inferMemberBindingsViaProperty(providedBindings, parameterType));

    default:
      throw new StateError('unsupported bind mode: ${bindMode}');
  }
}

_ModeBindingPair _inferBindingsViaConstructorThenProperty(
    Map<Symbol, Symbol> providedBindings, ClassMirror parameterType,
    Symbol constructorName) {

  final viaCtr = _inferMemberBindingsViaConstructor(providedBindings,
      parameterType, constructorName, throwOnMissingCtr: false);

  final bindMode = viaCtr != null ? BindMode.CONSTRUCTOR: BindMode.PROPERTY;

  final inferredBindings = viaCtr != null ? viaCtr :
    _inferMemberBindingsViaProperty(providedBindings, parameterType);

  return new _ModeBindingPair(bindMode, inferredBindings);
}


Map<Symbol, Symbol> _inferMemberBindingsViaConstructor(
    Map<Symbol, Symbol> providedBindings, ClassMirror parameterType,
    Symbol constructorName, {throwOnMissingCtr: true}) {
  final _providedBindings = new BiMap()..addAll(providedBindings);

  final fullCtrName = new Symbol("${MirrorSystem.getName(parameterType.simpleName)}"
    ".${MirrorSystem.getName(constructorName)}");

  final MethodMirror ctrMirror = parameterType.declarations[fullCtrName];
  if (ctrMirror == null) {
    if (throwOnMissingCtr) {
      throw new ArgumentError('No such constructor $fullCtrName');
    }
    else {
      return null;
    }
  }

  if (!ctrMirror.isConstructor) {
    throw new ArgumentError('$fullCtrName is not a constructor');
  }

  final memberBindings = <Symbol, Symbol>{};

  ctrMirror.parameters.forEach((pm) {
    final parameterName = pm.simpleName;
    final providedName = _providedBindings.inverse[parameterName];

    final requestName = providedName != null ? providedName : parameterName;
    memberBindings[requestName] = parameterName;
  });

  return memberBindings;
}

Map<Symbol, Symbol> _inferMemberBindingsViaProperty(
    Map<Symbol, Symbol> providedBindings, ClassMirror parameterType) {

  final _providedBindings = new BiMap()..addAll(providedBindings);

  final memberBindings = <Symbol, Symbol>{};

  parameterType.declarations.forEach((s, dm) {
    if (dm is VariableMirror || (dm is MethodMirror && dm.isGetter)) {
      final propertyName = dm.simpleName;
      final providedName = _providedBindings.inverse[propertyName];

      final requestName = providedName != null ? providedName : propertyName;
      memberBindings[requestName] = propertyName;
    }
  });

  return memberBindings;
}


TypeBinding _createTypeMapping(Type type, Symbol constructorName,
                                      Map<String, Symbol> memberBindings,
                                      BindMode bindMode) {
  return new TypeBinding(type, constructorName: constructorName,
      bindMode: bindMode, memberBindings: memberBindings);
}


InstanceMirror _createInstance(Request request, TypeBinding binding) {
  switch (binding.bindMode) {
    case BindMode.CONSTRUCTOR:
      return _createInstanceViaConstructor(request, binding);
    case BindMode.PROPERTY:
      return _createInstanceViaProperty(request, binding);

    default:
      throw new StateError('unsupported bind mode: ${binding.bindMode}');
  }
}

InstanceMirror _createInstanceViaConstructor(Request request,
                                             TypeBinding binding) {
  final namedArguments = _bindVariables(request, binding.memberBindings);
  final classMirror = reflectClass(binding.type);
  final instanceMirror =
      classMirror.newInstance(binding.constructorName, [], namedArguments);
  return instanceMirror;
}

InstanceMirror _createInstanceViaProperty(Request request,
                                          TypeBinding binding) {
  final properties = _bindVariables(request, binding.memberBindings);
  final classMirror = reflectClass(binding.type);
  final instanceMirror =
      classMirror.newInstance(const Symbol(''), []);

  properties.forEach((property, value) {
    instanceMirror.setField(property, value);
  });
  return instanceMirror;
}


Map<Symbol, dynamic> _bindVariables(Request request,
    Map<Symbol, Symbol> memberBindings) {
  final pv = getPathParameters(request);

  final convertedPv = <Symbol, Object>{};
  pv.forEach((k, v) {
    convertedPv[new Symbol(k)] = v;
  });

  final namedArguments = <Symbol, dynamic> {};
  memberBindings.forEach((k, v) {
    final value = convertedPv[k];
    if (value != null) {
      namedArguments[v] = value;
    }
  });

  return namedArguments;
}

dynamic _getAnnotation(DeclarationMirror dm, Type annotationType) {
  final resourceAnnotations = dm.metadata.where(
      (md) => reflectType(annotationType).isAssignableTo(md.type));
  if (resourceAnnotations.isEmpty) {
    return null;
  }

  return resourceAnnotations.first.reflectee;
}


_BoundHandler _wrapHandler(Function handler) {
  final ClosureMirror handlerMirror = reflect(handler);
  final functionMirror = handlerMirror.function;
  final params = functionMirror.parameters;
  final numParams = params.length;
  final returnType = functionMirror.returnType;

  return (List<dynamic> positionalArguments, Request request,
      ResponseBinding responseBinding) {
    final result = handlerMirror.apply(positionalArguments);

    final futureResponse = syncFuture(() => result.reflectee);

    return futureResponse.then((response) {
      if (response == null) {
        return new Response.internalServerError(body:
          'function ${functionMirror.simpleName} returned null');
      }
      return responseBinding.createResponse(request, reflect(response));
    });
  };
}

