// Copyright (c) 2014, The Shelf Bind project authors.
// Please see the AUTHORS file for details.
// All rights reserved. Use of this source code is governed by
// a BSD 2-Clause License that can be found in the LICENSE file.

library shelf_bind.core;

import 'package:shelf/shelf.dart';
import 'bindings.dart';
export 'bindings.dart';

/**
 * A Shelf Route HandlerAdapter for seamless integration
 */
Handler handlerAdapter(Function handler) => bind(handler);

/**
 * Bind properties (path parameters, body) of a [Request] to the parameters
 * of a [handler] function.
 *
 * Optionally the bindings can be specified by passing [parameterBindings].
 * The order of the [parameterBindings] must match the order of the parameters
 * to the function.
 *
 * When the [parameterBindings] are omitted the bindings are automatically
 * inferred.
 *
 * The inferrence supports all the features of the [HandlerParameterBinding]s
 * including:
 * * binding to request path parameters
 * * binding to json body
 * * binding via constructor
 * * binding via setters
 * * etc
 *
 * TODO: add examples and describe inferrence process
 */
Handler bind(Function handler,
      { List<HandlerParameterBinding> parameterBindings: const [],
        ResponseBinding responseBinding}) {
  return createHandlerBinding(handler, parameterBindings,
      responseBinding).handler;
}

/**
 * Binds a single request path parameter to a single parameter of the [handler]
 * function based on the type of the handler parameter.
 *
 * The binding will be inferred, however, the default inference can be overriden
 * by specifying any or all of the [type], [bindings], [bindMode],
 * [constructor] parameters.
 */
Handler bindRequestPathParameter(Function handler,
             { Type type, Map<String, Symbol> bindings,
               BindMode bindMode: BindMode.CONSTRUCTOR,
               Symbol constructor: #build }) {
  return bind(handler, parameterBindings:
        [createRequestParameterBinding(type, constructor, bindings,
                                  bindMode)]);
}



// TODO: delete & document breaking change
//Handler bind(Type type, Map<String, Symbol> bindings, Function handler,
//             { BindMode bindMode: BindMode.CONSTRUCTOR,
//               Symbol constructor: #build }) {
//  return new _RequestParamBinding(type, constructor, bindings, handler,
//      bindMode).handler;
//}

/**
 * Binds the body of the [Request], which must be in JSON format, to a single
 * parameter of the [handler] function based on the type of the handler
 * parameter.
 *
 * Optionally, the [type] of the function parameter can be specified. Otherwise
 * it is inferred
 */
Handler bindJsonBody(Function handler, {Type type}) {
  return bind(handler, parameterBindings:
        [createJsonBodyBinding(type)]);
}


// TODO: delete & document breaking change
//Handler bindJsonBody(Type type, Function handler) {
//  return new _JsonBodyBinding(type, handler).handler;
//}

